﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.BinPacking.Interfaces;

namespace HeuristicLab.Encodings.PackingEncoding.MultiComponentVector {
	[Item("Multi Component Vector Creator", "Creator class used to create MultiComponentVector solutions for bin packing problems.")]
	[StorableClass]
	public class MultiComponentVectorRandomCreator : PackingSolutionCreator, IStochasticOperator {

		public ILookupParameter<IRandom> RandomParameter {
			get { return (LookupParameter<IRandom>)Parameters["Random"]; }
		}
		public IValueLookupParameter<IntValue> PackingItemsParameter {
			get { return (IValueLookupParameter<IntValue>)Parameters["PackingItems"]; }
		}
		public IValueLookupParameter<IntValue> LowerBoundParameter {
			get { return (IValueLookupParameter<IntValue>)Parameters["LowerBound"]; }
		}
		public IValueParameter<BoolValue> SortedSequenceParameter {
			get { return (IValueParameter<BoolValue>)Parameters["SortedSequence"]; }
		}

		[StorableConstructor]
		protected MultiComponentVectorRandomCreator(bool deserializing) : base(deserializing) { }
		protected MultiComponentVectorRandomCreator(MultiComponentVectorRandomCreator original, Cloner cloner) : base(original, cloner) {
		}
		public override IDeepCloneable Clone(Cloner cloner) {
			return new MultiComponentVectorRandomCreator(this, cloner);
		}

		public MultiComponentVectorRandomCreator()
			: base() {
			Parameters.Add(new LookupParameter<IRandom>("Random", "The pseudo random number generator."));
			Parameters.Add(new ValueLookupParameter<IntValue>("PackingItems", "The number of packing-items handled in this problem instance."));
			Parameters.Add(new ValueLookupParameter<IntValue>("LowerBound", "The lower possible number of bins needed to solve this problem (taken from Dell'Amico, Martello and Vigo; 2002)"));
			Parameters.Add(new ValueParameter<BoolValue>("SortedSequence", "Determines if the items of the created instances are to be sorted by their size."));

			EncodedSolutionParameter.ActualName = "MultiComponentVector";
			SortedSequenceParameter.ActualValue = new BoolValue(false);
		}


		public static MultiComponentVectorEncoding Apply(int items, int lowerBound, IRandom random, bool sortedSequence) {
			var solution = new MultiComponentVectorEncoding();
			if (sortedSequence)
				solution.PackingInformations = PackingInformation.CreateDictionaryRandomlyWithSortedSequence(items, lowerBound, random);
			else
				solution.PackingInformations = PackingInformation.CreateDictionaryRandomly(items, lowerBound, random);
			return solution;
		}

		protected override IItem CreateSolution() {
			return Apply(PackingItemsParameter.ActualValue.Value, LowerBoundParameter.ActualValue.Value, RandomParameter.ActualValue, SortedSequenceParameter.Value.Value);
		}
	}
}
